from ..client.game_client import GameClient
from loguru import logger
import asyncio





class Map(GameClient):
    
    
    async def get_map_chunks_as(self, kingdom, x, y):
        
        try:
            
            await self.send_json_message("gaa",{
                "KID": kingdom,
                "AX1": x,
                "AY1": y,
                "AX2": x + 12,
                "AY2": y + 12
            })
            
            
            
        except Exception as e:
            logger.error(e)
            return False
    
    
    
    
    async def get_map_chunks_sync(
        self,
        kingdom: int,
        x: int,
        y: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "gaa",
                {
                    "KID": kingdom,
                    "AX1": x,
                    "AY1": y,
                    "AX2": x + 12,
                    "AY2": y + 12
                    
                }
            )
            if sync:
                response = await self.wait_for_response("gaa")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
    
    
    
    
    
    
    async def get_closest_npc(
        self,
        kingdom: int,
        npc_type: int,
        min_level: int = 1,
        max_level: int = -1,
        owner_id: int = -1,
        sync: bool = True
        ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "fnm",
                {
                    
                    "T": npc_type,
                    "KID": kingdom,
                    "LMIN": min_level,
                    "LMAX": max_level,
                    "NID": owner_id
                }
            )
            if sync:
                response = await self.wait_for_response("fnm")
                return response
            return True
            
        except Exception as e:
            logger.error(e)
            return False
        
        
    async def find_by_name(
        self,
        user_name: str,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("wsp", {"PN": user_name})
            if sync:
                response = await self.wait_for_response("wsp")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
    
    
    async def get_npc_target_infos(
        self,
        tx: int,
        ty: int,
        sx: int,
        sy: int,
        kid: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "adi",
                {
                    "TX": tx,
                    "TY": ty,
                    "SX": sx,
                    "SY": sy,
                    "KID": kid
                }
            )
            if sync:
                response = await self.wait_for_response("adi")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        




    async def get_pvp_target_infos(
        self,
        tx: int,
        ty: int,
        sx: int,
        sy: int,
        kid: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "aci",
                {
                    "TX": tx,
                    "TY": ty,
                    "SX": sx,
                    "SY": sy,
                    "KID": kid
                }
            )
            if sync:
                response = await self.wait_for_response("aci")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
    
    async def find_by_name(
        self,
        user_name: str,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("wsp", {"PN": user_name})
            if sync:
                response = await self.wait_for_response("wsp")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
    
    
    def _spiral_coords(self, max_radius: int, cx: int, cy: int):
        step = 13
        seen = set()
        for radius in range(step, max_radius + 1, step):

            for x in range(cx - radius, cx + radius + 1, step):
                pt = (x, cy - radius)
                if pt not in seen:
                    seen.add(pt)
                    yield pt
           
            for y in range(cy - radius, cy + radius + 1, step):
                pt = (cx + radius, y)
                if pt not in seen:
                    seen.add(pt)
                    yield pt
            
            for x in range(cx + radius, cx - radius - 1, -step):
                pt = (x, cy + radius)
                if pt not in seen:
                    seen.add(pt)
                    yield pt
           
            for y in range(cy + radius, cy - radius - 1, -step):
                pt = (cx - radius, y)
                if pt not in seen:
                    seen.add(pt)
                    yield pt
        
        seen.clear()
    

    
        
    async def map_scanner(
        self,
        kingdom: int,
        max_radius: int,
        castle_x: int,
        castle_y: int
        ):
        
        try:
            logger.info("Scanning...")
            for x, y in self._spiral_coords(max_radius=max_radius, cx=castle_x, cy=castle_y):
                await self.get_map_chunks_as(kingdom, x, y)
                await asyncio.sleep(0.01)
            logger.info("Complete!")
            
        except Exception as e:
            logger.error(e)
            


    async def map_multi_scanner(
        self,
        kingdom: int,
        max_radius: int,
        castle_x: int,
        castle_y: int,
        interval: float
        ) -> None:
        
        while True:
            
            await self.map_scanner(kingdom, max_radius, castle_x, castle_y)
            
            logger.info("Sleeping until start next scan", interval)
            await asyncio.sleep(interval)



## configurari de client ##
VERSION = "V 0.20"
GAME_VERSION_URL = "https://empire-html5.goodgamestudios.com/default/items/ItemsVersion.properties"
CLIENT_ORIGIN = "https://empire-html5.goodgamestudios.com"
SERVERS_DB = "https://empire-html5.goodgamestudios.com/config/network/1.xml"
LANG_DB = "https://langserv.public.ggs-ep.com/em/en"












### client headers ###
AD_HEADERS = {
    
        "Accept-Encoding": "gzip, deflate, br, zstd",
        "Accept-Language": "en-US,en;q=0.9",
        "Cache-Control": "no-cache",
        "Pragma": "no-cache",
        "DNT": "1",
        "Sec-GPC": "1"
}







DEFAULT_UA_LIST = [
    # — Browsere moderne (versiuni actualizate pentru HTML5/WebGL) —
    # Google Chrome pe Windows 11
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/120.0.0.0 Safari/537.36",

    # Google Chrome pe macOS Sonoma (14_0)
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 14_0) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/120.0.0.0 Safari/537.36",

    # Mozilla Firefox pe Windows 11
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:120.0) "
    "Gecko/20100101 Firefox/120.0",

    # Mozilla Firefox pe macOS Sonoma (14_0)
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 14_0; rv:120.0) "
    "Gecko/20100101 Firefox/120.0",

    # Microsoft Edge pe Windows 11
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/120.0.0.0 Safari/537.36 Edg/120.0.0.0",

    # Apple Safari pe macOS Sonoma (14_0)
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 14_0) "
    "AppleWebKit/605.1.15 (KHTML, like Gecko) "
    "Version/17.4 Safari/605.1.15",

    # Opera (Blink) pe Windows 11
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/120.0.0.0 Safari/537.36 OPR/103.0.4924.32",


    # — Intrări din DEFAULT_UA_LIST (fișier) —
    # Chrome pe Windows 10
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/115.0.5790.98 Safari/537.36",

    # Chrome pe macOS (10_15_7)
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/115.0.0.0 Safari/537.36",

    # Chrome pe Linux
    "Mozilla/5.0 (X11; Linux x86_64) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/114.0.0.0 Safari/537.36",

    # Firefox pe Windows
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:114.0) "
    "Gecko/20100101 Firefox/114.0",

    # Firefox pe macOS (10.15)
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:112.0) "
    "Gecko/20100101 Firefox/112.0",

    # Safari pe macOS (10_15_7)
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
    "AppleWebKit/605.1.15 (KHTML, like Gecko) "
    "Version/14.0 Safari/605.1.15",

    # Edge pe Windows
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/114.0.0.0 Safari/537.36 Edg/114.0.0.0",

    # Opera pe Windows
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/114.0.0.0 Safari/537.36 OPR/100.0.0.0",
]



from ..client.game_client import GameClient
from loguru import logger





class Social(GameClient):
    
    
    
    
    async def send_player_sms(
        self,
        player_name: str,
        sms_title: str,
        sms_text: str,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "sms",
                {
                    "RN": player_name,
                    "MH": sms_title,
                    "TXT": sms_text
                }
            )
            if sync:
                response = await self.wait_for_response("sms")
                return response
            return True

        except Exception as e:
            logger.error(e)
            
    
    
    async def read_messages(self, message_id: int, sync: bool = True):
        
        try:
            
            await self.send_json_message(
                "rms", {
                    "MID": message_id
                }
            )
            
            if sync:
                response = await self.wait_for_response("rms")
                return response
            return True
            
        except Exception as e:
            logger.error(e) 
            return False
    
    
    
    
    
    
    
            
    async def delete_message(self, message_id, sync = True):
        
        try:
            
            await self.send_json_message(
                "dms", {
                    "MID": message_id
                }
            )
            
            if sync:
                response = await self.wait_for_response("dms")
                return response
            logger.info(f'Message {message_id} removed!')
            return True
            
        except Exception as e:
            logger.error(e) 
            return False
        
    
    
    
    
    
    async def read_report(self, message_id: int, sync: bool = True):
        
        try:
            
            await self.send_json_message(
                "bsd", {
                    "MID": message_id
                }
            )
            
            if sync:
                response = await self.wait_for_response("bsd")
                return response
            
            return True
            
        except Exception as e:
            logger.error(e) 
            return False
        
        
    
    
    
    
    
    
    
        
        
    async def no_battle_handle(self, data):
        
        msg_data = data.get("MSG", [])
        for msg_detail in msg_data:
            if msg_detail[1] == 67:
                await self.delete_message(msg_detail[0])
    
            
    
    
    
    
    
    async def spy_report_handle(self, data):
        
        
        
        
        msg_data = data.get("MSG", [])
        for msg_detail in msg_data:
            if msg_detail[1] == 3:
                spy_check = str(msg_detail[2]).split('#', 1)[0]
                if spy_check.startswith("1+0"):
                    report = await self.read_report(msg_detail[0])
                    return report
                
                    




from ..client.game_client import GameClient
from loguru import logger




class Quests(GameClient):
    
    
    
    
    
    
    async def get_quests(
        self,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("dcl", {"CD": 1})
            if sync:
                response = await self.wait_for_response("dcl")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def complete_message_quest(
        self,
        quest_id: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("qsc", {"QID": quest_id})
            if sync:
                response = await self.wait_for_response("qsc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
     
     
        
        
    async def complete_donation_quest(
        self,
        quest_id: int,
        food: int = 0,
        wood: int = 0,
        stone: int = 0,
        gold: int = 0,
        oil: int = 0,
        coal: int = 0,
        iron: int = 0,
        glass: int = 0,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "qdr",
                {
                    "QID": quest_id,
                    "F": food,
                    "W": wood,
                    "S": stone,
                    "C1": gold,
                    "O": oil,
                    "C": coal,
                    "I": iron,
                    "G": glass,
                    "PWR": 0,
                    "PO": -1
                }
            )
            if sync:
                response = await self.wait_for_response("qdr")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def tracking_recommended_quests(self) -> bool:
        
        try:
            await self.send_json_message("ctr", {"TQR": 0})
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def complete_quest_condition(
        self,
        quest_id: int,
        condition: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "fcq",
                {
                    "QTID": quest_id,
                    "QC": condition
                }
            )
            if sync:
                response = await self.wait_for_response("fcq")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
 


from ..client.game_client import GameClient
from loguru import logger







class Spy(GameClient):
    
    
    
    
    async def send_spy(
        self,
        kingdom: int,
        source_id: int,
        tx: int,
        ty: int,
        spies_nr: int,
        precision: int,
        horses_type: int = -1,
        slowdown: int = 0,
        feathers: int = 0,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "csm",
                {
                    "SID": source_id,
                    "TX": tx,
                    "TY": ty,
                    "SC": spies_nr,
                    "ST": 0,
                    "SE": precision,
                    "HBW": horses_type,
                    "KID": kingdom,
                    "PTT": feathers,
                    "SD": slowdown
                }
            )
            if sync:
                response = await self.wait_for_response("csm")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def send_sabotage(
        self,
        kingdom: int,
        source_id: int,
        tx: int,
        ty: int,
        spies_nr: int,
        burn_pecent: int = 50,
        horses_type: int = 0,
        feathers: int = 0,
        slowdown: int = 0,
        sync: bool = True   
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "csm",
                {
                    "SID": source_id,
                    "TX": tx,
                    "TY": ty,
                    "SC": spies_nr,
                    "ST": 2,
                    "SE": burn_pecent,
                    "HBW": horses_type,
                    "KID": kingdom,
                    "PTT": feathers,
                    "SD": slowdown
                }
            )
            if sync:
                response = await self.wait_for_response("csm")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


import asyncio
import json
import random
import re
import inspect
from typing import Any, Callable, Dict, List, Optional, Union, Awaitable
import aiohttp
from loguru import logger
from websockets.asyncio.client import connect, ClientConnection
from websockets.exceptions import ConnectionClosed
from ._cfg import *


class GameClient:
    
    HandlerType = Callable[[Any], Union[Any, Awaitable[Any]]]
    
    def __init__(
        self,
        url: str,
        server_header: str,
        username: str,
        password: str
    ) -> None:
        
        self.url = url
        self.server_header = server_header
        self.username = username
        self.password = password
        
        self.ws: Optional[ClientConnection] = None
        self.connected = asyncio.Event()
        self._stop_event = asyncio.Event()
        self._tasks: List[asyncio.Task] = []
        
        self._msg_queue: asyncio.Queue[Dict[str, Any]] = asyncio.Queue()
        self._pending_futures: Dict[str, List[asyncio.Future]] = {}
        self.user_agent = random.choice(DEFAULT_UA_LIST)
        
    async def connect(self) -> None:
        """
        Main client loop to manage connection and reconnection.
        """
        while not self._stop_event.is_set():
            try:
                await self._run_connection_session()
            except ConnectionClosed:
                logger.warning("Connection closed, attempting to reconnect in 5 seconds...")
                self.connected.clear()
                await asyncio.sleep(5)
            except Exception as e:
                logger.error(f"Error during connection session: {e}, attempting to reconnect in 5 seconds...")
                self.connected.clear()
                await asyncio.sleep(5)
        logger.info("Client shutdown complete.")

    async def _run_connection_session(self) -> None:
        """
        Handles a single connection session.
        """
        async with connect(
            self.url,
            origin=CLIENT_ORIGIN,
            user_agent_header=self.user_agent,
            additional_headers=AD_HEADERS
        ) as ws:
            self.ws = ws
            self.connected.set()
            logger.info(f"GGClient connected! {VERSION}")
            
            listener = asyncio.create_task(self._listener())
            keep_alive = asyncio.create_task(self.keep_alive())
            nch_task = asyncio.create_task(self._nch())
            
            if not await self._init():
                await self.disconnect()
                return
            
            runner = asyncio.create_task(self.run_jobs())
            self._tasks = [listener, keep_alive, nch_task, runner]
            
            done, pending = await asyncio.wait(
                self._tasks,
                return_when=asyncio.FIRST_EXCEPTION
            )
            for t in pending: t.cancel()
            for t in done:
                if t.exception(): raise t.exception()
                
    async def _init(self) -> bool:
        await self._init_socket()
        return await self.login(self.username, self.password)            
                
    async def run_jobs(self) -> None:
        pass

    async def disconnect(self) -> None:
        self.connected.clear()
        if self.ws:
            try:
                await self.ws.close()
            except Exception as e:
                logger.error(f"Error closing connection: {e}")

        logger.info("Disconnected!.")
        
    async def shutdown(self) -> None:
        self._stop_event.set()
        await self.disconnect()
        
        for task in self._tasks:
            task.cancel()
        await asyncio.gather(*self._tasks, return_exceptions=True)
        
    async def send(self, message: str) -> None:
        if not self.ws: raise RuntimeError("GGClient not connected!")
        await self.ws.send(message)
            
    async def send_message(self, parts: List[str]) -> None:
        msg = "%".join(["", *parts, ""])
        await self.send(msg)
        
    async def send_raw_message(self, command: str, data: List[Any]) -> None:
        json_parts = [json.dumps(item) if isinstance(item, (dict, list)) else item for item in data]
        await self.send_message(["xt", self.server_header, command, "1", *json_parts])

    async def send_json_message(self, command: str, data: Dict[str, Any]) -> None:
        await self.send_message(["xt", self.server_header, command, "1", json.dumps(data)])

    async def send_xml_message(self, t: str, action: str, r: str, data: str) -> None:
        await self.send(f"<msg t='{t}'><body action='{action}' r='{r}'>{data}</body></msg>")
        
    async def receive(self) -> Dict[str, Any]:
        return await self._msg_queue.get()
    
    def _parse_message(self, message: str) -> Dict[str, Any]:
        if message.startswith("<"):
            m = re.search(r"<msg t='(.*?)'><body action='(.*?)' r='(.*?)'>(.*?)</body></msg>", message)
            t_val, action, r_val, data = m.groups()
            return {"type": "xml", "payload": {"t": t_val, "action": action, "r": int(r_val), "data": data}}
        
        parts = message.strip("%").split("%")
        cmd = parts[1]; status = int(parts[3])
        raw = "%".join(parts[4:])
        
        try:
            data = json.loads(raw)
        except:
            data = raw
        
        parsed_data = {"type": "json", "payload": {"command": cmd, "status": status, "data": data}}      
        return parsed_data
    
    async def _listener(self) -> None:
        try:
            async for raw in self.ws:
                text = raw.decode('utf-8') if isinstance(raw, bytes) else raw
                msg = self._parse_message(text)
                await self._msg_queue.put(msg)
                payload = msg.get("payload", {})
                cmd = payload.get("command") or payload.get("action")
                futures = self._pending_futures.get(cmd)
                if futures:
                    for fut in futures:
                        if not fut.done():
                            fut.set_result(payload.get("data"))
                    continue

                method = f"on_{cmd}"
                if hasattr(self, method):
                    handler = getattr(self, method)
                    data = payload.get("data")
                    if inspect.iscoroutinefunction(handler):
                        asyncio.create_task(handler(data))
                    else:
                        handler(data)
        except ConnectionClosed:
            logger.warning("Connection closed, stopping listener...")
            # Am eliminat apelul recursiv la self.connect() de aici.
            # Buclei principale de conectare ii va reveni responsabilitatea reconectarii.
            
    async def wait_for_response(self, command: str, timeout: float = 5.0) -> Any:
        deadline = asyncio.get_event_loop().time() + timeout
        buffer: List[Dict[str, Any]] = []
        try:
            while True:
                remaining = deadline - asyncio.get_event_loop().time()
                if remaining <= 0:
                    raise asyncio.TimeoutError(f"Timeout waiting for {command}")
                msg = await asyncio.wait_for(self._msg_queue.get(), timeout=remaining)
                payload = msg.get("payload", {})
                cmd = payload.get("command") or payload.get("action")
                msg_status = payload.get("status")
                if cmd == command and msg_status != 1:
                    return payload.get("data")
                buffer.append(msg)
        finally:
            for m in buffer:
                await self._msg_queue.put(m)
            
    async def send_rpc(self, command: str, data: Dict[str, Any], timeout: float = 5.0) -> Any:
        await self.send_json_message(command, data)
        return await self.wait_for_response(command, timeout)
    
    async def send_hrpc(self, command: str, data: Dict[str, Any], handler: HandlerType, timeout: float = 5.0) -> Any:
        await self.send_json_message(command, data)
        resp_data = await self.wait_for_response(command, timeout)
        to_handle = handler(resp_data)
        if inspect.isawaitable(to_handle):
            await to_handle
            
    async def keep_alive(self, interval: int = 60) -> None:
        await self.connected.wait()
        while self.connected.is_set() and not self._stop_event.is_set():
            await asyncio.sleep(interval)
            await self.send_raw_message("pin", ["<RoundHouseKick>"])
            
    async def _nch(self, interval: int = 360) -> None:
        await self.connected.wait()
        while self.connected.is_set():
            await asyncio.sleep(interval)
            await self.send(f'%xt%{self.server_header}%nch%1%')
        
    async def _init_socket(self):
        await self.send_xml_message("sys", "verChk", "0", "<ver v='166' />")
        await self.send_xml_message("sys", "login", "0", 
                                        f"<login z='{self.server_header}'><nick><![CDATA[]]></nick><pword><![CDATA[1123010%fr%0]]></pword></login>")
        await self.send_xml_message("sys", "autoJoin", "-1", "")
        await self.send_xml_message("sys", "roundTrip", "1", "")
            
    async def fetch_game_db(self) -> dict:
        async with aiohttp.ClientSession() as session:
            async with session.get(GAME_VERSION_URL) as resp:
                resp.raise_for_status()
                text = await resp.text()
                _, version = text.strip().split("=", 1)
                version = version.strip()
            
            db_url = f"https://empire-html5.goodgamestudios.com/default/items/items_v{version}.json"
            async with session.get(db_url) as db_resp:
                db_resp.raise_for_status()
                data = await db_resp.json()
                return data
            
    async def login(
        self,
        username: str,
        password: str
        ) -> bool:
        
        if not self.connected.is_set():
            logger.error("Not connected yet!")
            
        try:
            await self.send_json_message(
                "lli",
                {
                    "CONM": 175,
                    "RTM": 24,
                    "ID": 0,
                    "PL": 1,
                    "NOM": username,
                    "PW": password,
                    "LT": None,
                    "LANG": "fr",
                    "DID": "0",
                    "AID": "1674256959939529708",
                    "KID": "",
                    "REF": "https://empire.goodgamestudios.com",
                    "GCI": "",
                    "SID": 9,
                    "PLFID": 1
                }
                )
            
            response = await self.wait_for_response("lli")
            
            if not isinstance(response, dict):
                return True
            
            if isinstance(response, dict) and not response:
                logger.warning("Wrong username or password!")
                return False
            
            if isinstance(response, dict):
                cooldown_value = response["CD"]
                logger.debug(f'Connection locked by the server! Reconnect in {cooldown_value} sec!')
                await asyncio.sleep(cooldown_value)
                return await self.login(username, password)
                 
        except Exception as e:
            logger.error(e)
            return False
        
        logger.error("Unexpected response in login status!")
        return False


from ..client.game_client import GameClient
from loguru import logger





class Defense(GameClient):
    
    
    
    async def get_castle_defense(
        self, x: int, y: int, castle_id: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "dfc", {"CX": x, "CY": y, "AID": castle_id, "KID": -1, "SSV": 0}
            )
            if sync:
                response = await self.wait_for_response("dfc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            
            
    async def change_keep_defnse(
        self,
        x: int,
        y: int,
        castle_id: int,
        min_units_to_consume_tools: int,
        melee_percentage: int,
        tools: list[list[int]],
        support_tools: list[list[int]],
        sync: bool = True
        
            
    ) -> dict | bool:
        
        try:
            
            await self.send_json_message(
                "dfk",
                {
                    "CX": x,
                    "CY": y,
                    "AID": castle_id,
                    "MAUCT": min_units_to_consume_tools,
                    "UC": melee_percentage,
                    "S": tools,
                    "STS": support_tools,
                }
            )
            if sync:
                response = await self.wait_for_response("dfk")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            
            
    async def change_wall_defense(
        self,
        x: int,
        y: int,
        castle_id: int,
        left_tools: list[list[int]],
        left_unit_percentage: int,
        left_melee_percentage: int,
        middle_tools: list[list[int]],
        middle_unit_percentage: int,
        middle_melee_percentage: int,
        right_tools: list[list[int]],
        right_unit_percentage: int,
        right_melee_percentage: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            
            await self.send_json_message(
                "dfw",
                {
                    "CX": x,
                    "CY": y,
                    "AID": castle_id,
                    "L": {
                        "S": left_tools,
                        "UP": left_unit_percentage,
                        "UC": left_melee_percentage
                    },
                    "M": {
                        "S": middle_tools,
                        "UP": middle_unit_percentage,
                        "UC": middle_melee_percentage
                    },
                    "R": {
                        "S": right_tools,
                        "UP": right_unit_percentage,
                        "UC": right_melee_percentage,
                    }
                }
            )
            if sync:
                response = await self.wait_for_response("dfw")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            
            
            
    async def change_moat_defense(
        self,
        x: int,
        y: int,
        castle_id: int,
        left_tools: list[list[int]],
        middle_tools: list[list[int]],
        right_tools: list[list[int]],
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "dfm",
                {
                    "CX": x,
                    "CY": y,
                    "AID": castle_id,
                    "LS": left_tools,
                    "MS": middle_tools,
                    "RS": right_tools
                }
            )
            if sync:
                response = await self.wait_for_response("dfm")
                return response
            return True
        
        except Exception as e:
            logger.error(e)


from ..client.game_client import GameClient
from loguru import logger





class Hospital(GameClient):
    
    
    
    
    async def heal(
        self, wod_id: int, amount: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("hru", {"U": wod_id, "A": amount})
            if sync:
                response = await self.wait_for_response("hru")
                return response
            return True
        
        
        except Exception as e:
            logger.error(e)
            
            
    async def cancel_heal(
        self, slot_id: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("hcs", {"S": slot_id})
            if sync:
                response = self.wait_for_response("hcs")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            
            
    async def skip_heal(
        self, slot_id: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            self.send_json_message("hss", {"S": slot_id})
            if sync:
                response = await self.wait_for_response("hss")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            
            
    
    async def delete_wounded(
        self, wod_id: int, amount: int, sync: bool = True
    )-> dict | bool:
        
        try:
            await self.send_json_message("hdu", {"U": wod_id, "A": amount})
            if sync:
                response = await self.wait_for_response("hdu")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            
            
    async def ask_alliance_help_heal(
        self, package_id: int, sync: bool = True
    )-> dict | bool:
        
        try:
            await self.send_json_message("ahr", {"ID": package_id, "T": 2})
            if sync:
                response = await self.wait_for_response("ahr")
                return response
            return True
        
        except Exception as e:
            logger.error(e)


from ..client.game_client import GameClient
from loguru import logger





class Lords(GameClient):
    
    
    async def get_lords(self, sync = True) -> dict | bool:
        
        
        try:
            
            await self.send_json_message("gli", {})
            
            if sync:
                response = await self.wait_for_response("gli")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
    
    

    
    
    
    
    
    async def select_lord(
        self,
        lord_index: int,
        lord_list: list,    
    ) -> int:
        
        if not hasattr(self, 'lord_index'):
            self.lord_index = 0
            
        if not isinstance(lord_list, list):
            logger.error("No selected lords available!")
            return None
        
        n = len(lord_list)
        for _ in range(n):
            lord = lord_list[lord_index]
            lord_index = (lord_index + 1) % n
            return lord
        
        
        
        
    async def list_lords_id(
        self,
        lord_list: list,
    ) -> list:
        
        if not isinstance(lord_list, list):
            lord_list = []
        
        lords_data = await self.get_lords()
        all_lords = lords_data.get("C", [])
        
        for lord_obj in all_lords:
            lord_id = lord_obj.get("ID")
            eq_list = lord_obj.get("EQ")
            if len(eq_list) >= 5:
                lord_list.append(lord_id)
        
        


from ..client.game_client import GameClient
from loguru import logger




class OuterRealms(GameClient):
    
    
    async def get_outer_realms_points(self, sync: bool = True) -> dict | bool:
        
        try:
            await self.send_json_message("tsh", {})
            if sync:
                response = await self.wait_for_response("tsh")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
        
    async def choose_outer_realms_castle(
        self,
        castle_id: int,
        only_rubies: int = 0,
        use_rubies: int = 0,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "tsc",
                {
                    "ID": castle_id,
                    "OC2": only_rubies,
                    "PWR": use_rubies,
                    "GST": 3
                }
            )
            if sync:
                response = await self.wait_for_response("tsc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def get_outer_realms_token(self, sync: bool = True) -> dict | bool:
        
        try:
            await self.send_json_message(
                "glt",
                {
                    "GST": 3
                }
            )
            if sync:
                response = await self.wait_for_response("glt")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def login_outer_realms(
        self,
        token: str,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "tlep",
                {
                    "TLT": token
                }
            )
            if sync:
                response = await self.wait_for_response("lli")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger




class BuildingsInventory(GameClient):
    
    
    
    async def get_building_inventory(
        self, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("sin", {})
            if sync:
                response = await self.wait_for_response("sin")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def store_building(
        self, building_id: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("sob", {"OID": building_id})
            if sync:
                response = await self.wait_for_response("sob")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
    async def sell_building_inventory(
        self,
        wod_id: int,
        amount: int,
        unique_id: int = -1,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "sds", {"WID": wod_id, "AMT": amount, "UID": unique_id}
            )
            if sync:
                response = await self.wait_for_response("sds")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger
from ..utils.utils import Utils
import asyncio





## de adaugat soldier_replesnished cu cap








class Castle(GameClient):
    
    
    
    async def go_to_kingdom(
        self,
        loc_id: int,
        kid: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "jca",
                {
                    "CID": loc_id,
                    "KID": kid
                }
            )
            if sync:
                response = await self.wait_for_response("jaa")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
    
    
    
    async def go_to_mains(
        self,
        px: int,
        py: int,
        kid: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "jaa",
                {
                    "PX": px,
                    "PY": py,
                    "KID": kid
                }
            )
            if sync:
                response = await self.wait_for_response("jaa")
                return response
            return True

        except Exception as e:
            logger.error(e)
            return False
    
    
    
    
    
    
    async def get_castles(self, sync: bool = True) -> dict | bool:
        
        try:
            
            await self.send_json_message("gcl", {})
            if sync:
                response = await self.wait_for_response("gcl")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def get_detailed_castles(self, sync: bool = True) -> dict | bool:
        
        try:
            
            await self.send_json_message("dcl", {})
            if sync:
                response = await self.wait_for_response("dcl")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
        
        
        
        
    async def send_resources_to_kingdom(
        self,
        id_sender: int,
        sender_kid: int,
        target_kid: int,
        resources: list[list[str, int]],
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "kgt",
                {
                   "SCID": id_sender,
                   "SKID": sender_kid,
                   "TKID": target_kid,
                   "G": resources
                }
            )
            if sync:
                response = await self.wait_for_response("kgt")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
    async def send_units_to_kingdom(
        self,
        id_sender: int,
        sender_kid: int,
        target_kid: int,
        units: list[list[int, int]],
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "kut",
                {
                   "SCID": id_sender,
                   "SKID": sender_kid,
                   "TKID": target_kid,
                   "CID": -1,
                   "A": units
                }
            )
            if sync:
                response = await self.wait_for_response("kut")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def skip_kingdom_transfer(
        self,
        skip: str,
        target_kid: int,
        transfer_type: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "msk",
                {
                    "MST": skip,
                    "KID": target_kid,
                    "TT": transfer_type
                }
            )
            if sync:
                response = await self.wait_for_response("msk")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def auto_units_kingdom_transfer(
        self,
        id_sender: int,
        sender_kid: int,
        target_kid: int,
        units: list,
        skips: list = None,
        sync: bool = True  
    ) -> dict | bool:
        
        
        utils = Utils()
        send_units = await self.send_units_to_kingdom(id_sender, sender_kid, target_kid, units)
        if not isinstance(send_units, dict):
            logger.error(f"Failed to send units to kingdom: {target_kid}")
            return
        
        kpi = send_units.get("kpi", {})
        ut_list = kpi.get("UT")
        if not ut_list or not isinstance(ut_list, list):
            logger.error("Unknown transfer time data!")
            return
        
        time_to_transfer = ut_list[0].get("RS")
        if not isinstance(time_to_transfer, int):
            logger.error(f"Invalid time value: {ut_list[0]}")
            return
        
        skip_list = utils.skip_calculator(time_to_transfer, skips)
        for skip in skip_list:
            await self.skip_kingdom_transfer(skip, target_kid, transfer_type=1, sync=sync)
        
        logger.info(f"All units has been sent succesfully to kingdom {target_kid}!")
        
        
        
        
        
        
    async def auto_res_kingdom_transfer(
        self,
        id_sender: int,
        sender_kid: int,
        target_kid: int,
        resources: list,
        skips: list = None,
        sync: bool = True  
    ) -> dict | bool:
        
        
        utils = Utils()
        resources_sender = await self.send_resources_to_kingdom(id_sender, sender_kid, target_kid, resources)
        if not isinstance(resources_sender, dict):
            logger.error(f"Failed to send resources to kingdom: {target_kid}")
            return
        
        kpi = resources_sender.get("kpi", {})
        rt_list = kpi.get("RT")
        if not rt_list or not isinstance(rt_list, list):
            logger.error("Unknown transfer time data!")
            return
        
        time_to_transfer = rt_list[0].get("RS")
        if not isinstance(time_to_transfer, int):
            logger.error(f"Invalid time value: {rt_list[0]}")
            return

        skip_list = utils.skip_calculator(time_to_transfer, skips)
        for skip in skip_list:
            await self.skip_kingdom_transfer(skip, target_kid, transfer_type=2, sync=sync)
        
        logger.info(f"All resources has been sent succesfully to kingdom {target_kid}!")    
        
        
        
    async def kingdom_auto_feeder(
        self,
        target_kid: int,
        activator: bool,
        min_food: int,
        min_mead: int,
        skips: list = None,
        interval: float = 60.0,
        sync: bool = True
    ) -> None:
        
        
        while activator:
            
            try:
                castles_inventory = await self.get_detailed_castles(sync = sync)
                resource_inventory = castles_inventory["C"]
                
                potential_donors_food = {}
                potential_donors_mead = {}
                
            except TimeoutError:
                await asyncio.sleep(1)
                continue
            
            for items in resource_inventory:
                kid = items.get("KID")
                if kid == target_kid:
                    continue
                
                for ai_item in items.get("AI", []):
                    aid = ai_item.get("AID")
                    mead_value = ai_item.get("MEAD")
                    food_value = ai_item.get("F")
                    mead_prod = ai_item.get("gpa", {}).get("DMEAD")
                    food_prod = ai_item.get("gpa", {}).get("DF")
                    
                    if food_value > 100000 and food_prod > 0:
                        potential_donors_food[kid] = {"aid": aid, "food": food_value}
                    
                    if mead_value > 100000 and mead_prod > 0:
                        potential_donors_mead[kid] = {"aid": aid, "mead": mead_value}
                        
            
            for items in resource_inventory:
                kid = items.get("KID")
                if kid == target_kid:
                    for ai_item in items.get("AI", []):
                        food_value = ai_item.get("F")
                        mead_value = ai_item.get("MEAD")
                        food_cap = ai_item.get("gpa", {}).get("MRF")
                        mead_cap = ai_item.get("gpa", {}).get("MRMEAD")
                        fcap_var = int(food_cap - 5) - int(food_value)
                        mcap_var = int(mead_cap - 5) - int(mead_value)
                        
                        
                        if food_value < min_food:
                            logger.warning(f"Kingdom {kid} has low level of food: {food_value}. Searching for food...")
                            if potential_donors_food:
                                donor_kid = max(potential_donors_food, key=lambda k: potential_donors_food[k]["food"])
                                donor_info = potential_donors_food[donor_kid]
                                donor_aid = donor_info["aid"]
                                logger.info(f"Transfer from kingdom {donor_kid} (AID: {donor_aid}) to kingdom {kid}.")
                                await self.auto_res_kingdom_transfer(donor_aid, donor_kid, target_kid, [["F", fcap_var]], skips, sync)
                            
                            else:
                                logger.warning("No enough amount of food in your castles to refill!")
                        
                        
                        if mead_value < min_mead:
                            logger.warning(f"Kingdom {kid} has low level of mead: {mead_value}. Searching for...")
                            if potential_donors_mead:
                                donor_kid = max(potential_donors_mead, key=lambda k: potential_donors_mead[k]["mead"])
                                donor_info = potential_donors_mead[donor_kid]
                                donor_aid = donor_info["aid"]
                                logger.info(f"Transfer mead from kingdom {donor_kid} (AID: {donor_aid}) to kingdom {kid}.")
                                await self.auto_res_kingdom_transfer(donor_aid, donor_kid, target_kid, [["MEAD", mcap_var]], skips, sync)
                                
                            else:
                                logger.warning("No enough amount of mead in your castles to refill!")
            
            await asyncio.sleep(interval)
            
            
            
            
    async def units_replenish(
        self,
        target_kid: int,
        wod_id: int,
        amount: int
    ) -> None:
        
        
        try:
            
            account_inventory = await self.get_detailed_castles()
            inventory_data = account_inventory["C"]
            donors = []
            
            
            for kingdom in inventory_data:
                kid = kingdom.get("KID")
                if kid == target_kid:
                    continue
                
                for ai_block in kingdom.get("AI", []):
                    aid = ai_block.get("AID")
                    for wod, amt in ai_block.get("AC", []):
                        if wod == wod_id and amt > amount:
                            donors.append({"aid": aid, "kid":kid, "amount": amt})
                            break
                        
            if not donors:
                logger.warning("I can't find any eligible location!")
                return False
                            
            else:
                best = max(donors, key=lambda d: d["amount"])
                donor_aid = best["aid"]
                donor_amt = best["amount"]
                donor_kid = best["kid"]
                send_amt = min(donor_amt, amount)
                
                await self.auto_units_kingdom_transfer(donor_aid, donor_kid, target_kid, [[wod_id, send_amt]])
                logger.info(f"Kingdom {target_kid} refilled with {send_amt} units!")
                     
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger




class Movements(GameClient):
    
    
    async def get_movements(self, sync = True) -> dict | bool:
        
        try:
            
            await self.send_json_message("gam", {})
            
            if sync:
                response = await self.wait_for_response("gam")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


    async def retrieve_army(
        self,
        movement_id: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("mcm", {"MID": movement_id})
            if sync:
                response =  await self.wait_for_response("mcm")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from setuptools import setup, find_packages

setup(
    name="ggx",
    version="0.3.1",
    description="API for GoodGame Empire",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="mst4ck",
    author_email="mst4ck@mailfence.com",
    license="MIT",
    packages=find_packages(),      
    install_requires=[
        "websockets>=15.0.1",
        "loguru>=0.7.3",
        "aiohttp>=3.12.13",
    ],
    python_requires=">=3.12",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



from ..client.game_client import GameClient
from loguru import logger





class Repair(GameClient):
    
    
    
    async def repair_building(
        self, building_id: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("rbu", {"OID": building_id, "PO": -1, "PWR": 0})
            if sync:
                response = await self.wait_for_response("rbu")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            
            
    async def ask_alliance_help_repair(
        self, building_id: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("ahr", {"ID": building_id, "T": 3})
            if sync:
                response = await self.wait_for_response("ahr")
                return response
            return True
        except Exception as e:
            logger.error(e)
        


from ..client.game_client import GameClient
from loguru import logger






class Events(GameClient):
    
    
    async def get_events(self, sync: bool = True) -> dict | bool:
        
        try:
            await self.send_json_message("sei", {})
            if sync:
                response = await self.wait_for_response("sei")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def get_event_points(
        self,
        event_id: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "pep",
                {
                    "EID": event_id
                }
            )
            if sync:
                response = await self.wait_for_response("pep")
                return response
            return True

        except Exception as e:
            logger.error(e)
            return False       
        
        
    async def get_ranking(
        self,
        ranking_type: int,
        category: int = -1,
        search_value: int = -1,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "hgh",
                {
                    "LT": ranking_type,
                    "LID": category,
                    "SV": search_value
                }
            )
            if sync:
                response = await self.wait_for_response("hgh")
                return response
            return True

        except Exception as e:
            logger.error(e)
            return False  
        
        
        
    async def choose_event_difficulty(
        self,
        event_id: int,
        difficulty_id: int,
        premium_unlock: int = 0,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "sede",
                {
                    "EID": event_id,
                    "EDID": difficulty_id,
                    "C2U": premium_unlock
                }
            )
            if sync:
                response = await self.wait_for_response("sede")
                return response
            return True

        except Exception as e:
            logger.error(e)
            return False 